require "quaternion"
require "model"
require "spoon"

Scene = Object:new()

local hw, hh = love.graphics.getWidth()/2, love.graphics.getHeight()/2

function Scene:Scene(vfov)
	self.focallen = .5/math.tan(vfov/2)
	self.pos = Vector:new(0, 0, 0)
	self.rot = Quaternion(0, 0, 0, 1);
	self.mdl = Model:new(spoon, Vector:new(0, 0, 4))

	love.mouse.setGrab(true)
	love.mouse.setVisible(false)
end

function Scene:keypressed(key)
	if key == "escape" then
      	love.event.push("quit")
   	end
end

function Scene:update(dt)
	--print(1/dt)
	local mx, my = love.mouse.getX(), love.mouse.getY()

	self.rot = self.rot*Quaternion:fromAngle(0, 0, -((mx/hw)-1))
	self.rot = Quaternion:fromAngle(0, ((my/hh)-1), 0)*self.rot
	self.rot:normal()

	if love.keyboard.isDown("w") then
		local move = self.rot:rotate(Vector:new(0, 0, 3*dt))
		self.pos = self.pos - move 
	end

	self.mdl.rot = self.mdl.rot*Quaternion:fromAngle(1*dt, 0, 1*dt)

	love.mouse.setPosition(hw, hh)
end

function Scene:draw()
	self:drawmdl(self.mdl)
end

local nverts, sverts = {}, {}

local tempvec = Vector:new(0, 0, 0)
local avg = Vector:new(0, 0, 0)

-- Draw a model
function Scene:drawmdl(mdl)
	love.graphics.push()
	love.graphics.translate(hw, hh)
	love.graphics.scale(love.graphics.getWidth())

	local depthn = 0
	local flen, rot = self.focallen, self.rot
	local mrot = mdl.rot
	local mpos = self.pos + mdl.pos

	if not nverts[mdl] then nverts[mdl] = {} sverts[mdl] = {} end
	for k,v in pairs(mdl.vertices) do
		if not nverts[k] then nverts[k] = Vector:new() sverts[k] = {0, 0} end
		tempvec.x, tempvec.y, tempvec.z = mrot:inlinerotate(v)
		nverts[k].x, nverts[k].y, nverts[k].z = rot:inlinerotate(tempvec:add(mpos))

		sverts[k].x = nverts[k].x * (flen/nverts[k].z)
		sverts[k].y = nverts[k].y * (flen/nverts[k].z)

		if nverts[k].z < 0 then sverts[k].x = math.huge sverts[k].y = math.huge end
	end

	for k,v in pairs(mdl.faces) do
		love.graphics.setColor((150-math.abs(sverts[v[1]].x*500))%255, 30, (150-math.abs(sverts[v[1]].x*500))%255)
		love.graphics.triangle("fill", 
			sverts[v[1]].x, sverts[v[1]].y,
			sverts[v[2]].x, sverts[v[2]].y,
			sverts[v[3]].x, sverts[v[3]].y
		)
	end

	love.graphics.pop()
end